// Builder logic v2 

#include "Hitters.as";
#include "Hitters2.as";
#include "Knocked.as";
#include "BuilderCommon.as";
#include "ThrowCommon.as";
#include "RunnerCommon.as";
#include "MakeMat.as";
#include "Help.as";
#include "Requirements.as";
#include "BuilderHittable.as";
#include "PlacementCommon.as";

bool cooldown;
u8 timer;
//can't be <2 - needs one frame less for gathering infos
const s32 hit_frame = 2;
f32 hit_damage = 0.5f;

void onInit( CBlob@ this )
{
    this.set_f32( "pickaxe_distance", 10.0f );
    //no spinning
    this.getShape().SetRotationsAllowed(false);
    this.set_f32("gib health", -3.0f);
    this.Tag("player");
    this.Tag("flesh");
    HitData hitdata;
    this.set("hitdata", hitdata );
    this.addCommandID("pickaxe");
    this.getShape().getConsts().net_threshold_multiplier = 0.5f;
    this.set_u8("axeTimer", 0);
    this.set_bool("axeCooldown", false);
    cooldown = false;
    timer = 0;
    this.set_u8("toolNumber", 0);
    this.set_u8("toolTimer", 0);
    this.set_u8("cooldownTimer", 0);
    this.set_bool("toolCooldown", false);

    this.set_Vec2f("tileAim", Vec2f_zero);

    this.set_Vec2f("inventory offset", Vec2f(0.0f, 160.0f));

    SetHelp( this, "help self action2", "builder", "$Pick$Dig/Chop  $KEY_HOLD$$RMB$", "", 3 );

    this.getCurrentScript().runFlags |= Script::tick_not_attached;
    this.getCurrentScript().removeIfTag = "dead";
}

void onSetPlayer( CBlob@ this, CPlayer@ player )
{
    if (player !is null)
    {
        player.SetScoreboardVars("ScoreboardIcons.png", 1, Vec2f(16,16));
    }
}

void onTick(CBlob@ this)
{
    if (this.isInInventory())
        return;

    const bool ismyplayer = this.isMyPlayer();

    if (ismyplayer && getHUD().hasMenus()) {
        return;
    }   

    u8 toolNumber = this.get_u8("toolNumber");
    setupTool(this, toolNumber, this.isKeyPressed(key_action2));//tool functionality all in one function.

    if(ismyplayer) //client
    {
        
    }

        if (this.isKeyJustPressed(key_action3))
        {
            CBlob@ carried = this.getCarriedBlob();
            if (carried is null || !carried.hasTag("temp blob")) {
                client_SendThrowOrActivateCommand( this );
            }
        }



    if (this.isKeyPressed(key_action1) && this.get_u8("toolNumber") != 1)
    {
        BlockCursor @bc;
        this.get( "blockCursor", @bc );

        HitData@ hitdata;
        this.get("hitdata", @hitdata);
        hitdata.blobID = 0;
        hitdata.tilepos = bc.buildable ? bc.tileAimPos : Vec2f(-8,-8);
    }

    // get rid of the built item

    if (this.isKeyJustPressed(key_inventory) || this.isKeyJustPressed(key_pickup))
    {
        this.set_u8( "buildblob", 255 );
        this.set_TileType( "buildtile", 0 );
        CBlob@ blob = this.getCarriedBlob();
        if (blob !is null && blob.hasTag("temp blob"))
        {
            blob.Untag("temp blob");
            blob.server_Die();
        }
    }
}

void setupTool(CBlob@ this, u8 toolNumber, bool key)
{
    if(toolNumber < tools.length)
    {
        u8 timer = this.get_u8("toolTimer");
        u8 hitTick = tools[toolNumber].hitTick;
        u8 coolTime = tools[toolNumber].coolTime;
        u8 coolTimer = this.get_u8("cooldownTimer");
        bool cooldown = this.get_bool("toolCooldown");

        if (cooldown)
        {
            coolTimer--;
            this.set_u8("cooldownTimer", coolTimer);
        }
        if (coolTimer <= 0)
        {
            this.set_u8("cooldownTimer", 0);
            this.Sync("cooldownTimer", false);
            this.set_bool("toolCooldown", false);
            this.Sync("toolCooldown", false);
        }

        if(key)
        {
            if(!cooldown)
            {
                timer++;
                this.set_u8("toolTimer", timer);
                //print(timer+"");
                if (timer >= hitTick)
                {
                    this.set_bool("hit!", true);
                    this.Sync("hit!", false);
                    this.set_u8("toolTimer", 0);
                    this.Sync("toolTimer", false);
                }
            }
        }
        else if(!key && timer < hitTick)
        {
            this.set_u8("toolTimer", 0);
            this.Sync("toolTimer", false);
        }
        if(this.get_bool("hit!") == true)
        {
            switch(toolNumber)
            {
                case 0:
                    Pickaxe(this);
                    break;
                case 1:
                    Axe(this, 0.5f, Hitters2::axe);
                    break;
            }
            this.set_u8("cooldownTimer", coolTime);
            this.Sync("cooldownTimer", false);
            this.set_bool("toolCooldown", true);
            this.Sync("toolCooldown", false);
            this.set_bool("hit!", false);
            this.Sync("hit!", false);
        }
    }
    else
    {
        print("invalid tool! void setupTool in builderlogic.as");
    }
}

void SendHitCommand( CBlob@ this, CBlob@ blob, const Vec2f tilepos, const Vec2f attackVel, const f32 attack_power )
{
    CBitStream params;
    if (blob is null)
        params.write_netid( 0 );
    else
        params.write_netid( blob.getNetworkID() );
    params.write_Vec2f( tilepos );
    params.write_Vec2f( attackVel );
    params.write_f32( attack_power );
    this.SendCommand( this.getCommandID("pickaxe"), params );
}

bool RecdHitCommand( CBlob@ this, CBitStream@ params )
{
    u16 blobID;
    Vec2f tilepos, attackVel;
    f32 attack_power;
    if (!params.saferead_netid( blobID ))
        return false;
    if (!params.saferead_Vec2f( tilepos ))
        return false;
    if (!params.saferead_Vec2f( attackVel ))
        return false;
    if (!params.saferead_f32( attack_power ))
        return false;

    if (blobID == 0)
    {
        CMap@ map = getMap();
        if(map !is null)
        {
            bool noBuildZone = map.getSectorAtPosition( tilepos, "no build") !is null;
            if(!noBuildZone)
            {
                this.server_HitMap( tilepos, attackVel, 1.0f, Hitters::builder );
            }
        }
    }
    return true;
}

void onCommand( CBlob@ this, u8 cmd, CBitStream @params )
{
    if (cmd == this.getCommandID("pickaxe"))
    {
        if (!RecdHitCommand( this, params ))
            warn("error when recieving pickaxe command");
    }
}

class SortHitsParams {
    Vec2f aimPos;
    Vec2f tilepos;
    Vec2f pos;
    bool hasHit;
    HitInfo@ bestinfo;
    f32 bestDistance;
};

void Pickaxe(CBlob@ this)
{
    if (!getNet().isServer()) {
        return;
    }

    HitData@ hitdata;
    if (!this.get("hitdata", @hitdata)) {
        return;
    }

    Vec2f blobPos = this.getPosition();
    Vec2f aimPos = this.getAimPos();
    Vec2f aimDir = aimPos - blobPos;

    // get tile surface for aiming at little static blobs
    Vec2f normal = aimDir;
    normal.Normalize();
    Vec2f attackVel = normal;

    f32 aimangle = aimDir.Angle();
    Vec2f pos = blobPos - Vec2f(2,0).RotateBy(-aimangle);
    f32 attack_distance = this.getRadius() + 10.0f;


    f32 radius = this.getRadius();
    CMap@ map = this.getMap();
    f32 arcdegrees = 90.0f;
    bool hasHit = false;
    HitInfo@ bestinfo = null;
    f32 bestDistance = 100000.0f;

    const f32 tile_attack_distance = attack_distance * 1.5f;
    Vec2f tilepos = blobPos + normal * Maths::Min(aimDir.Length() - 1, tile_attack_distance);

    bool noBuildZone = map.getSectorAtPosition( tilepos, "no build") !is null;
    bool tilehit = false;

    hitdata.tilepos = Vec2f_zero;

    Vec2f surfacepos;
    map.rayCastSolid( blobPos, tilepos, surfacepos );
    hitdata.tilepos = surfacepos;

    if( (tilepos - aimPos).Length() < bestDistance && map.getBlobAtPosition(tilepos) is null )
    {
        Tile tile = map.getTile( surfacepos );

        if (!noBuildZone && !map.isTileGroundBack( tile.type ))
        {
            //normal, honest to god tile
            if(map.isTileBackgroundNonEmpty( tile ) || map.isTileSolid( tile ) || map.isTileGrass( tile.type ) )
            {
                if(hitdata.tilepos != Vec2f_zero) //error checking, don't hit topleft of map.
                {
                    hitdata.blobID = 0;
                    SendHitCommand( this, null, hitdata.tilepos, attackVel, hit_damage );
                    if(!map.isTileGrass( tile.type ))
                    {
                    	tilehit = true;
                    }
                }
                else
                {
                    print("tiles are dumb");
                }
            }
        }
    }
    //handling blobs sucks.
    //Let's get the distance between the mouse and the player.
    f32 distance = (blobPos - aimPos).Length();
    Vec2f area;
    if(distance > attack_distance)
    {
    	area = blobPos + (normal * attack_distance);
    }
    else
    {
    	area = aimPos;
    }
    if(!tilehit)//dont hit blobs if hitting tiles
    {
        HitInfo@[] hitInfos;
        if (map.getHitInfosFromArc( area, -aimangle, 360, 7.0f, this, @hitInfos ))
        {
            HitInfo@ hit = hitInfos[0]; //Hit closest blob to mouse position

            CBlob@ b = hit.blob;
            HitInfo@[] hitInfos2;
            if(map.getHitInfosFromArc( blobPos, -aimangle, 18, attack_distance, this, @hitInfos2 ))
            {
            	//RayCastSolid doesn't work for blob tiles, use some dumb workaround logic.
			    for (uint i = 0; i < hitInfos2.length; i++)
			    {
			    	HitInfo@ trace = hitInfos2[i];
			    	CBlob@ t = trace.blob;
				    if(t !is null)
				    {
				    	if(t.getShape().isStatic())
				    	{
				    		const bool teamHurt = (!t.hasTag("flesh") || t.hasTag("dead"));
			                this.server_Hit( t, t.getPosition(), attackVel, 0.5f, Hitters::builder, teamHurt);
			                break;
				    	}
				    	else if (b !is null && canHit( this, b, b.getPosition(), false))
			            {
			                const bool teamHurt = (!b.hasTag("flesh") || b.hasTag("dead"));
			                this.server_Hit( b, tilepos, attackVel, 0.5f, Hitters::builder, teamHurt);
			                break;
			            }
		        	}
			    }
            }
        }
    }
}

bool canHit( CBlob@ this, CBlob@ b, Vec2f tpos, bool extra = true )
{

    if(b.hasTag("dead"))
        return true;

    if(b.hasTag("invincible"))
        return false;

    //team stuff
    if(b.getTeamNum() == this.getTeamNum())
    {
        //
        if(b.hasTag("flesh") || b.hasTag("player"))
        {
            return false;
        }

        //static or background stuff
        CShape@ bshape = b.getShape(); 
        if(!b.isCollidable() || (bshape !is null && bshape.isStatic()))
        {
            //maybe we shouldn't hit this..
            //check if we should always hit
            return BuilderAlwaysHit(b);
        }
    }

    //should be hit
    return true;
    
}


void Axe(CBlob@ this, f32 damage, u8 type)
{
    if (!getNet().isServer()) {
        return;
    }
    
    f32 aimangle = (this.isFacingLeft() ? 180 : 0 );
    CSprite@ sprite = this.getSprite();
    bool justCheck = !sprite.isFrameIndex(hit_frame);
    bool adjusttime = sprite.getFrameIndex() < hit_frame-1;

    Vec2f blobPos = this.getPosition();
    Vec2f pos = blobPos - Vec2f(6,0).RotateBy(aimangle);
    f32 attack_distance = 18.0f;
    f32 radius = this.getRadius();
    CMap@ map = this.getMap();
    //get the actual aim angle
    f32 exact_aimangle = (this.getAimPos() - blobPos).Angle();
    bool dontHitMore = false;

    // this gathers HitInfo objects which contain blob or tile hit information
    HitInfo@[] hitInfos;
    if (map.getHitInfosFromArc( blobPos, -exact_aimangle, 50.0f, radius + attack_distance, this, @hitInfos) )
    {
        //HitInfo objects are sorted, first come closest hits
        for (uint i = 0; i < hitInfos.length; i++)
        {
            HitInfo@ hi = hitInfos[i];

            CBlob@ b = hi.blob;

            if (b !is null && !dontHitMore) // hit blobs, not tiles
            {
                print(""+b.getName());
                //big things block attacks
                const bool large = b.hasTag("blocks sword") && !b.isAttached() && b.isCollidable();
                
                if (b.getTeamNum() == this.getTeamNum() && !b.hasTag("dead"))
                { // no TK
                    if(large)
                        dontHitMore = true;
                    
                    continue;
                }
                Vec2f velocity = b.getPosition() - pos;
                this.server_Hit( b, hi.hitpos, velocity, damage, type, true); // server_Hit() is server-side only

            }
            else
            {
                Vec2f position = this.getPosition();
                f32 scalar = 16.0f;

                Vec2f target = position - hi.hitpos;
                target.Normalize();
                target *= -scalar;

                Vec2f end;
                //print(" position = "+position.x+", "+position.y);
                //print(" aim = "+this.getAimPos().x+", "+this.getAimPos().y);
                //print(" target = "+target.x+", "+target.y);

                if(!map.rayCastSolidNoBlobs(position, position + target, end)) return;
                TileType tile = map.getTile(end).type;

                if(!map.isTileWood(tile)) return;

                map.server_DestroyTile(end, 1.0f);
                map.server_DestroyTile(end, 1.0f);
            }
            
        }
    }
}

void onDetach( CBlob@ this, CBlob@ detached, AttachmentPoint@ attachedPoint )
{
    // ignore collision for built blob
    BuildBlock[]@ blocks;
    this.get( "blocks", @blocks );  
    for (uint i = 0; i < blocks.length; i++)
    {
        BuildBlock@ b = blocks[i];
        if (b.name == detached.getName())
        {
            this.IgnoreCollisionWhileOverlapped( null );
            detached.IgnoreCollisionWhileOverlapped( null );
        }
    }

    // BUILD BLOB
    // take requirements from blob that is built and play sound
    // put out another one of the same
    if (detached.hasTag("temp blob"))
    {
        if (!detached.hasTag("temp blob placed"))
        {
            detached.server_Die();
            return;
        }       

        uint i = this.get_u8( "buildblob" );
        if (blocks !is null && i >= 0 && i < blocks.length)
        {
            BuildBlock@ b = blocks[i];  
            if (b.name == detached.getName())
            {
                CInventory@ inv = this.getInventory();
                CBitStream missing;   
                this.set_u8( "buildblob", 255 );
                this.set_TileType( "buildtile", 0 );
                if (hasRequirements( inv, b.reqs, missing ))
                {
                    server_TakeRequirements( inv, b.reqs );
                }
                // take out another one if in inventory
                server_BuildBlob( this, @blocks, i );

            }
        }
    }
    else // take out another seed
        if (detached.getName() == "seed")
        {
            CBlob@ anotherBlob = this.getInventory().getItem( detached.getName() );

            if (anotherBlob !is null)   {
                this.server_Pickup( anotherBlob );
            }
        }
}

void onAddToInventory( CBlob@ this, CBlob@ blob )
{
    // destroy built blob if somehow they got into inventory
    if (blob.hasTag("temp blob"))
    {
        blob.server_Die();
        blob.Untag("temp blob");
    }

    if (this.isMyPlayer() && blob.hasTag("material"))
    {       
        SetHelp( this, "help inventory", "builder", "$Help_Block1$$Swap$$Help_Block2$           $KEY_HOLD$$KEY_F$", "", 3 );        
    }
}